import numpy as np
from tqdm import tqdm
from ..utils.utils import Hbeta, pca

def x2p(X, distance_matrix=False, tol=1e-5, perplexity=30.0):
    # Initialize some variables
    print("Computing pairwise distances...")
    (n, d) = X.shape
    
    if not distance_matrix:
        sum_X = np.sum(np.square(X), 1)
        D = np.add(np.add(-2 * np.dot(X, X.T), sum_X).T, sum_X)
    else:
        D = X

    P = np.zeros((n, n))
    beta = np.ones((n, 1))
    logU = np.log(perplexity)

    # Loop over all datapoints 
    for i in tqdm(range(n)):

        # Compute the Gaussian kernel and entropy for the current precision
        betamin = -np.inf
        betamax = np.inf
        Di = D[i, np.concatenate((np.r_[0:i], np.r_[i+1:n]))]
        (H, thisP) = Hbeta(Di, beta[i])

        # Evaluate whether the perplexity is within tolerance
        Hdiff = H - logU
        tries = 0
        while np.abs(Hdiff) > tol and tries < 50:

            # If not, increase or decrease precision
            if Hdiff > 0:
                betamin = beta[i].copy()
                if betamax == np.inf or betamax == -np.inf:
                    beta[i] = beta[i] * 2.
                else:
                    beta[i] = (beta[i] + betamax) / 2.
            else:
                betamax = beta[i].copy()
                if betamin == np.inf or betamin == -np.inf:
                    beta[i] = beta[i] / 2.
                else:
                    beta[i] = (beta[i] + betamin) / 2.

            # Recompute the values
            (H, thisP) = Hbeta(Di, beta[i])
            Hdiff = H - logU
            tries += 1

        # Set the final row of P
        P[i, np.concatenate((np.r_[0:i], np.r_[i+1:n]))] = thisP

    # Return final P-matrix
    print("Mean value of sigma: %f" % np.mean(np.sqrt(1 / beta)))
    return P


def jedi(X=np.array([]),Z=np.array([]), alpha=0.5, beta=0.5, n_components=2, perplexity=30.0, n_iter=1000, dens=False, verbose=1, random_seed=None, initial_dims=None):

    if random_seed:
        np.random.seed(random_seed)

    # Check inputs
    if not isinstance(n_components, int):
        print("Error: array X should have type float.")
        return -1

    # Initialize variables
    if initial_dims:
        X = pca(X, initial_dims).real
    (n, d) = X.shape
    
    initial_momentum = 0.5
    final_momentum = 0.8
    eta = 500
    min_gain = 0.01
    Y = np.random.randn(n, n_components)
    dY_KL = np.zeros((n, n_components))
    dY_JS = np.zeros((n, n_components))
    iY = np.zeros((n, n_components))
    gains = np.ones((n, n_components))

    def KL_divergence(P, Q):
        return np.sum(P * np.log(P / Q))


    def cost_function(P, Q, P_prime, alpha, beta):
        JS_alpha_beta = alpha * KL_divergence(P_prime, beta * Q + (1 - beta) * P_prime) + (1 - alpha) * KL_divergence(Q, beta * P_prime + (1 - beta)* Q)
        return KL_divergence(P, Q) - JS_alpha_beta


    def KL_divergence_grad_update(P, Q): 
        PQ = P - Q
        for i in range(n):
            dY_KL[i, :] = (PQ[i, :] * num[i, :]) @ (Y[i, :] - Y)
            
    
    def JS_alpha_beta_grad_update(P, Q, alpha, beta, num): 
        common_vec1 = P*Q / (beta*Q + (1-beta)*P)
        common_vec1[range(n), range(n)] = 0
        common_vec1 = alpha*beta*common_vec1.sum()

        common_vec2 = Q*(1 + np.log(Q) - (1-beta)*Q / (beta*P + (1-beta)*Q) - np.log(beta*P + (1-beta)*Q))
        common_vec2[range(n), range(n)] = 0
        common_vec2 = (1-alpha) * common_vec2.sum()

        for i in range(n):
            vec_before_brackets = num[i, :] * Q[i, :] 
            vec1 = alpha * beta * P[i, :] / (beta * Q[i, :] + (1 - beta) * P[i, :])
            vec2 = (1-alpha)*(-1-np.log(Q[i, :]) + (1-beta)*Q[i, :]/(beta*P[i, :]+(1-beta)*Q[i, :]) + np.log(beta*P[i, :]+(1-beta)*Q[i, :]))
            vec = vec_before_brackets * (vec1 - common_vec1 + vec2 + common_vec2)   
            dY_JS[i, :] = vec @ (Y[i, :] - Y) # sum over j
        

    # Compute matrix P
    P = x2p(X, distance_matrix=False, tol=1e-5, perplexity=perplexity)
    P = P + np.transpose(P)
    P = P / np.sum(P)
    # Early exaggeration
    P = P * 4.			
    P = np.maximum(P, 1e-12)

    # Compute matrix P_prime 
    P_prime = x2p(Z, distance_matrix=True, tol=1e-5, perplexity=perplexity)
    P_prime = P_prime + np.transpose(P_prime)
    P_prime = P_prime / np.sum(P_prime)
    # Early exaggeration
    P_prime = P_prime * 4.			
    P_prime = np.maximum(P_prime, 1e-12)
    
    C_old = np.sum(P)+100
    # Run iterations
    print('Performing optimization...')
    for iter in tqdm(range(n_iter)):

        # Compute pairwise affinities
        sum_Y = np.sum(np.square(Y), 1)
        num = -2. * np.dot(Y, Y.T)
        num = 1. / (1. + np.add(np.add(num, sum_Y).T, sum_Y))
        num[range(n), range(n)] = 0.
        Q = num / np.sum(num)
        Q = np.maximum(Q, 1e-12)

        # Compute gradients
        KL_divergence_grad_update(P, Q)
        JS_alpha_beta_grad_update(P_prime, Q, alpha, beta, num)

        # Perform the update
        if iter < 20:
            momentum = initial_momentum
        else:
            momentum = final_momentum
        gains = (gains + 0.2) * (((dY_KL+dY_JS) > 0.) != (iY > 0.)) + \
                (gains * 0.8) * (((dY_KL+dY_JS) > 0.) == (iY > 0.))
        gains[gains < min_gain] = min_gain
        iY = momentum * iY - eta * (gains * (dY_KL+dY_JS))
        Y = Y + iY
        Y = Y - np.tile(np.mean(Y, 0), (n, 1))

        # Compute current value of cost function
        if (iter + 1) % 50 == 0 and verbose > 1:
            C = np.sum(P * np.log(P / Q))
            if abs(C-C_old)<5e-5 and iter > 100:
                print("Early stopping due to small change", C, C_old)
                break
            C_old = C
            print("Iteration %d: error is %f" % (iter + 1, C))

        # Stop lying about P-values
        if iter == 100:
            P = P / 4.

    # Return solution
    return Y




import numpy as np
from tqdm import tqdm
from ..utils.utils import Hbeta, pca



def BetaDens(D=np.array([]), beta=np.array([]), i=None):
    """
        Compute the perplexity and the P-row for a specific value of the
        precision of a Gaussian distribution.
    """

    # Compute P-row and corresponding perplexity
    n = D.shape[0]
    #symm_beta = 1. / ((1. / np.sqrt(beta[np.concatenate((np.r_[0:i], np.r_[i+1:n+1]))] / 2.) + 1. / np.sqrt(beta[i] / 2.)) / 2.)**2 / 2.
    symm_beta = 4 * beta * beta[i] / (2 * np.sqrt(beta * beta[i]) + beta[i] + beta)
    gamma = symm_beta.copy()
    #print(beta[i], symm_beta[i], gamma[i], beta.mean(), symm_beta.mean(), gamma.mean())
    P = np.exp(-D.copy() * symm_beta[np.concatenate((np.r_[0:i], np.r_[i+1:n+1])),0])
    sumP = max(sum(P),np.finfo(np.double).eps)
    P = P / sumP
    return P, gamma[np.concatenate((np.r_[0:i], np.r_[i+1:n+1])),0]

def x2p(X=np.array([]), tol=1e-5, perplexity=30.0, dens=False):
    """
        Performs a binary search to get P-values in such a way that each
        conditional Gaussian has the same perplexity.
    """

    # Initialize some variables
    print("Computing pairwise distances...")
    (n, d) = X.shape
    sum_X = np.sum(np.square(X), 1)
    D = np.add(np.add(-2 * np.dot(X, X.T), sum_X).T, sum_X)
    P = np.zeros((n, n))
    gamma = np.zeros((n, n))
    beta = np.ones((n, 1))
    logU = np.log(perplexity)

    # Loop over all datapoints
    for i in tqdm(range(n)):

        # Compute the Gaussian kernel and entropy for the current precision
        betamin = -np.inf
        betamax = np.inf
        Di = D[i, np.concatenate((np.r_[0:i], np.r_[i+1:n]))]
        (H, thisP) = Hbeta(Di, beta[i])
        # Evaluate whether the perplexity is within tolerance
        Hdiff = H - logU
        tries = 0
        while np.abs(Hdiff) > tol and tries < 50:

            # If not, increase or decrease precision
            if Hdiff > 0:
                betamin = beta[i].copy()
                if betamax == np.inf or betamax == -np.inf:
                    beta[i] = beta[i] * 2.
                else:
                    beta[i] = (beta[i] + betamax) / 2.
            else:
                betamax = beta[i].copy()
                if betamin == np.inf or betamin == -np.inf:
                    beta[i] = beta[i] / 2.
                else:
                    beta[i] = (beta[i] + betamin) / 2.

            # Recompute the values
            (H, thisP) = Hbeta(Di, beta[i])
            Hdiff = H - logU
            tries += 1

        # Set the final row of P
        P[i, np.concatenate((np.r_[0:i], np.r_[i+1:n]))] = thisP
    if dens:
        for i in range(n):
            Di = D[i, np.concatenate((np.r_[0:i], np.r_[i+1:n]))]
            thisP, thisgamma = BetaDens(Di, beta, i)
            P[i, np.concatenate((np.r_[0:i], np.r_[i+1:n]))] = thisP
            gamma[i, np.concatenate((np.r_[0:i], np.r_[i+1:n]))] = thisgamma
        gamma /= gamma.max()
    # Return final P-matrix
    print("Mean value of sigma: %f" % np.mean(np.sqrt(1 / beta)))
    return P, gamma


def dtsne(X=np.array([]), n_components=2, perplexity=30.0, n_iter=1000, dens=False, verbose=1, random_seed=None, initial_dims=None):
    
    if random_seed:
        np.random.seed(random_seed)
    # Check inputs
    if not isinstance(n_components, int):
        print("Error: array X should have type float.")
        return -1
    if dens:
        print("Runs dtSNE variant of algorithm.")
    # Initialize variables
    if initial_dims:
        X = pca(X, initial_dims).real
    (n, d) = X.shape
    initial_momentum = 0.5
    final_momentum = 0.8
    eta = 500
    min_gain = 0.01
    Y = pca(X, n_components).real
    Y = 1e-4 * Y / np.std(Y)
    dY = np.zeros((n, n_components))
    iY = np.zeros((n, n_components))
    gains = np.ones((n, n_components))

    # Compute P-values
    P, gamma = x2p(X, 1e-5, perplexity, dens)
    P = P + np.transpose(P)
    P = P / np.sum(P)
    # Early exaggeration
    P = P * 4.  
    P = np.maximum(P, np.finfo(np.double).eps)
    
    C_old = np.sum(P)+100
    # Run iterations
    print('Performing optimization...')
    for iter in tqdm(range(n_iter)):

        # Compute pairwise affinities
        sum_Y = np.sum(np.square(Y), 1)
        num = -2. * np.dot(Y, Y.T)
        if dens:
            num = 1. / (1. + gamma * np.add(np.add(num, sum_Y).T, sum_Y))
        else:
            num = 1. / (1. + np.add(np.add(num, sum_Y).T, sum_Y))
        num[range(n), range(n)] = 0.
        Q = num / np.sum(num)
        Q = np.maximum(Q, 1e-12)

        # Compute gradient
        PQ = P - Q
        if dens:
            for i in range(n):
                dY[i, :] = 4 * np.sum(np.tile(PQ[:, i] * num[:, i] * gamma[:, i], (n_components, 1)).T * (Y[i, :] - Y) , 0)
        else:
            for i in range(n):
                dY[i, :] = np.sum(np.tile(PQ[:, i] * num[:, i], (n_components, 1)).T * (Y[i, :] - Y), 0)

        # Perform the update
        if iter < 20:
            momentum = initial_momentum
        else:
            momentum = final_momentum
        gains = (gains + 0.2) * ((dY > 0.) != (iY > 0.)) + \
                (gains * 0.8) * ((dY > 0.) == (iY > 0.))
        gains[gains < min_gain] = min_gain
        iY = momentum * iY - eta * (gains * dY)
        Y = Y + iY
        #Y = Y - np.tile(np.mean(Y, 0), (n, 1))

        # Compute current value of cost function
        
        if (iter + 1) % 50 == 0 and verbose > 1:
            C = np.sum(P * np.log(P / Q))
            if abs(C-C_old)<5e-5 and iter > 100:
                print("Early stopping due to small change", C, C_old)
                break
            C_old = C
            print("Iteration %d: error is %f" % (iter + 1, C))

        # Stop lying about P-values
        if iter == 100:
            P = P / 4.
    
    # Return solution
    return Y


def get_gaussian_data(dims=2, n_clusters=5):
    X = []
    y = []
    from scipy.stats import ortho_group

    for cluster in range(n_clusters):
        means = 100 * (np.random.rand(dims) - 0.5)
        diag = np.abs(np.diag(40 * np.random.rand(dims)))
        O = ortho_group.rvs(dims)
        covs = O.T @ diag @ O
        num_samples = np.random.randint(10, 200)

        cloud = np.random.multivariate_normal(means, covs, num_samples)

        X.append(cloud)
        y.append([cluster] * num_samples)

    return np.concatenate(X), np.concatenate(y)
    



from scipy.stats import ortho_group
import numpy as np
def get_gaussian_data(dims=2, n_clusters=5):
    X = []
    y = []

    for cluster in range(n_clusters):
        means = 100 * (np.random.rand(dims) - 0.5)
        diag = np.abs(np.diag(40 * np.random.rand(dims)))
        O = ortho_group.rvs(dims)
        covs = O.T @ diag @ O
        num_samples = np.random.randint(10, 200)

        cloud = np.random.multivariate_normal(means, covs, num_samples)

        X.append(cloud)
        y.append([cluster] * num_samples)

    return np.concatenate(X), np.concatenate(y)

def get_gaussian_data_jedi(dims=2, n_clusters=5):
    X = []
    y = []
    cluster_size = []
    from scipy.stats import ortho_group

    for cluster in range(n_clusters):
        means = 100 * (np.random.rand(dims) - 0.5)
        diag = np.abs(np.diag(40 * np.random.rand(dims)))
        O = ortho_group.rvs(dims)
        covs = O.T @ diag @ O
        num_samples = np.random.randint(100,300)

        cloud = np.random.multivariate_normal(means, covs, num_samples)

        X.append(cloud)
        y.append([cluster] * num_samples)
        cluster_size.append(num_samples)
    return np.concatenate(X), np.concatenate(y), cluster_size


import numpy as np
from scipy import stats as sps
import matplotlib.pyplot as plt


def rho(pts1, pts2):
    assert len(pts1.shape) == 2, "pts1 must be 2-d array"
    assert len(pts2.shape) == 2, "pts2 must be 2-d array"
    assert pts1.shape[0] == pts2.shape[0], "arrays pts1, pts2 must have the same number of points"
    n_points = pts1.shape[0]
    dim1 = pts1.shape[1]
    dim2 = pts2.shape[1]
    dists1 = np.zeros((n_points, n_points, dim1))
    dists2 = np.zeros((n_points, n_points, dim2))

    dists1 += pts1[:, None, :]
    dists1 -= pts1[None, :, :]
    dists1 = np.linalg.norm(dists1, axis=-1)

    dists2 += pts2[:, None, :]
    dists2 -= pts2[None, :, :]
    dists2 = np.linalg.norm(dists2, axis=-1)

    stat, _ = sps.spearmanr(dists1, dists2, axis=None)
    return stat


def rho_r(pts1, pts2, k=100):
    assert len(pts1.shape) == 2, "pts1 must be 2-d array"
    assert len(pts2.shape) == 2, "pts2 must be 2-d array"
    assert pts1.shape[0] == pts2.shape[0], "arrays pts1, pts2 must have the same number of points"
    n_points = pts1.shape[0]
    k = min(k, n_points-1)
    dim1 = pts1.shape[1]
    dim2 = pts2.shape[1]
    dists1 = np.zeros((n_points, n_points, dim1))
    dists2 = np.zeros((n_points, n_points, dim2))

    dists1 += pts1[:, None, :]
    dists1 -= pts1[None, :, :]
    dists1 = np.linalg.norm(dists1, axis=-1)

    dists2 += pts2[:, None, :]
    dists2 -= pts2[None, :, :]
    dists2 = np.linalg.norm(dists2, axis=-1)

    sorted_dists1 = np.sort(dists1, axis=0)
    sorted_dists2 = np.sort(dists2, axis=0)
    radii1 = sorted_dists1[k]
    radii2 = sorted_dists2[k]

    stat, _ = sps.spearmanr(radii1, radii2, axis=None)
    return stat


def rho_knn(pts1, pts2, k=100):
    assert len(pts1.shape) == 2, "pts1 must be 2-d array"
    assert len(pts2.shape) == 2, "pts2 must be 2-d array"
    assert pts1.shape[0] == pts2.shape[0], "arrays pts1, pts2 must have the same number of points"
    n_points = pts1.shape[0]
    k = min(k, n_points-1)
    dim1 = pts1.shape[1]
    dim2 = pts2.shape[1]
    dists1 = np.zeros((n_points, n_points, dim1))
    dists2 = np.zeros((n_points, n_points, dim2))

    dists1 += pts1[:, None, :]
    dists1 -= pts1[None, :, :]
    dists1 = np.linalg.norm(dists1, axis=-1)

    dists2 += pts2[:, None, :]
    dists2 -= pts2[None, :, :]
    dists2 = np.linalg.norm(dists2, axis=-1)

    sorted_dists1 = np.sort(dists1, axis=0)
    sorted_dists2 = np.sort(dists2, axis=0)
    k_nearest_neigbours1 = sorted_dists1[1:k+1]
    k_nearest_neigbours2 = sorted_dists2[1:k+1]

    stat, _ = sps.spearmanr(k_nearest_neigbours1,
                            k_nearest_neigbours2,
                            axis=None)
    return stat


def reconstruction_quality(pts1, pts2, k=100, plot=False):
    assert len(pts1.shape) == 2, "pts1 must be 2-d array"
    assert len(pts2.shape) == 2, "pts2 must be 2-d array"
    assert pts1.shape[0] == pts2.shape[0], "arrays pts1, pts2 must have the same number of points"
    n_points = pts1.shape[0]
    k = min(k, n_points-1)
    dim1 = pts1.shape[1]
    dim2 = pts2.shape[1]
    dists1 = np.zeros((n_points, n_points, dim1))
    dists2 = np.zeros((n_points, n_points, dim2))

    dists1 += pts1[:, None, :]
    dists1 -= pts1[None, :, :]
    dists1 = np.linalg.norm(dists1, axis=-1)

    dists2 += pts2[:, None, :]
    dists2 -= pts2[None, :, :]
    dists2 = np.linalg.norm(dists2, axis=-1)

    sorted_dists1 = np.sort(dists1, axis=0)
    sorted_dists2 = np.sort(dists2, axis=0)
    k_nearest_neigbours1 = sorted_dists1[1:k+1]
    k_nearest_neigbours2 = sorted_dists2[1:k+1]
    radii1 = sorted_dists1[k]
    radii2 = sorted_dists2[k]

    rho, _ = sps.spearmanr(dists1, dists2, axis=None)
    rho_knn, _ = sps.spearmanr(k_nearest_neigbours1,
                               k_nearest_neigbours2,
                               axis=None)
    rho_r, _ = sps.spearmanr(radii1, radii2, axis=None)
    if plot:
        plt.title("Reconstruction quality", fontsize=15)
        plt.bar([0, 1, 2], [rho, -rho_knn, rho_r], width=0.5,
                color=["r", "g", "b"],
                label=[r'$\rho$', r'$\rho_{knn}$', r'$\rho_{r}$'])
        plt.xlim(-0.75, 2.75)
        plt.ylim(-1, 1)
        plt.ylabel("Correlation")
        plt.xticks([0, 1, 2], ["global\nreconstruction",
                               "local\nreconstruction",
                               "relative\ndensity\nreconstruction"])
        plt.legend()
        plt.show()
    return rho, rho_knn, rho_r


def neighbourhood_overlap_score(pts1, pts2, plot=False):
    assert len(pts1.shape) == 2, "pts1 must be 2-d array"
    assert len(pts2.shape) == 2, "pts2 must be 2-d array"
    assert pts1.shape[0] == pts2.shape[0], "arrays pts1, pts2 must have the same number of points"
    
    def helper(l1, l2):
        number_of_common_neighbours = np.zeros(n_points)
        set1 = set()
        set2 = set()
        for k in range(1, n_points):
            number_of_common_neighbours[k] = number_of_common_neighbours[k-1]
            if k > 2:
                number_of_common_neighbours[k-1] /= (k-1)
            if l1[k] in set2:
                number_of_common_neighbours[k] += 1
            set1.add(l1[k])
            if l2[k] in set1:
                number_of_common_neighbours[k] += 1
            set2.add(l2[k])
        number_of_common_neighbours[-1] /= (n_points-1)
        return number_of_common_neighbours

    n_points = pts1.shape[0]
    dim1 = pts1.shape[1]
    dim2 = pts2.shape[1]
    dists1 = np.zeros((n_points, n_points, dim1))
    dists2 = np.zeros((n_points, n_points, dim2))

    dists1 += pts1[:, None, :]
    dists1 -= pts1[None, :, :]
    dists1 = np.linalg.norm(dists1, axis=-1)

    dists2 += pts2[:, None, :]
    dists2 -= pts2[None, :, :]
    dists2 = np.linalg.norm(dists2, axis=-1)
    
    sorted_dists1 = np.argsort(dists1, axis=0)
    sorted_dists2 = np.argsort(dists2, axis=0)
    
    nos = np.zeros((n_points, n_points))
    for i in range(n_points):
        nos[i] = helper(sorted_dists1[:, i], sorted_dists2[:, i])
    nos = nos.mean(axis=0)
    if plot:
        plt.title("Neighbourhood overlap score", fontsize=15)
        xs = np.linspace(0., 1., n_points)[1:]
        plt.plot(xs, nos[1:], label="NOS")
        plt.hlines(1, 0, 1, linestyle='--', color='black', alpha=0.5)
        plt.xlim(-0.05, 1.05)
        plt.ylim(-0.05, 1.05)
        plt.xlabel("Neighbourhood size in % of data")
        plt.ylabel("Neighbourhood overlap in %")
        plt.legend(loc="lower right")
        plt.show()
    return nos


if __name__=='__main__':
    data = np.random.rand(1500, 3)
    O = sps.ortho_group.rvs(3)
    transformed_data = data @ O
    rho, rho_knn, rho_r = reconstruction_quality(data, transformed_data)
    print('Calculated metrics:', rho, rho_knn, rho_r)
    assert np.isclose(rho, 1)
    assert np.isclose(rho_knn, 1)
    assert np.isclose(rho_r, 1)


import numpy as np

def pca(X=np.array([]), no_dims=50):
    """
        Runs PCA on the NxD array X in order to get initiall embedding.
    """

    print("Getting initall embedding using PCA...")
    (n, d) = X.shape
    X = X - np.tile(np.mean(X, 0), (n, 1))
    (l, M) = np.linalg.eig(np.dot(X.T, X))
    Y = np.dot(X, M[:, 0:no_dims])
    return Y

def Hbeta(D=np.array([]), beta=1.0):
    """
        Compute the perplexity and the P-row for a specific value of the
        precision of a Gaussian distribution.
    """

    # Compute P-row and corresponding perplexity
    P = np.exp(-D.copy() * beta) 
    sumP = max(sum(P),np.finfo(np.double).eps)
    H = np.log(sumP) + beta * np.sum(D * P) / sumP
    P = P / sumP
    return H, P



from setuptools import setup, find_packages
 
setup(name = "dtsnejedi",
      author="Aleksandr Odnakov",
      author_email="me@dnakov.ooo",
      packages = find_packages(),
      install_requires = ['matplotlib','numpy','scipy','tqdm'],
      version='0.1.1')